@@ -1,19 +1,25 @@
 import asyncio
 import os
+import pathlib
+import re
 import unittest
-from collections.abc import Sized, Container, Iterable
+from collections.abc import Sized, Container, Iterable, Mapping, MutableMapping
 from unittest import mock
 from urllib.parse import unquote
 import aiohttp.web
 from aiohttp import hdrs
 from aiohttp.web import (UrlDispatcher, Request, Response,
-                         HTTPMethodNotAllowed, HTTPNotFound)
+                         HTTPMethodNotAllowed, HTTPNotFound,
+                         HTTPCreated)
 from aiohttp.multidict import CIMultiDict
 from aiohttp.protocol import HttpVersion, RawRequestMessage
 from aiohttp.web_urldispatcher import (_defaultExpectHandler,
                                        DynamicRoute,
                                        PlainRoute,
-                                       SystemRoute)
+                                       SystemRoute,
+                                       ResourceRoute,
+                                       AbstractResource,
+                                       View)
 
 
 class TestUrlDispatcher(unittest.TestCase):
@@ -29,7 +35,7 @@ class TestUrlDispatcher(unittest.TestCase):
     def make_request(self, method, path):
         self.app = mock.Mock()
         message = RawRequestMessage(method, path, HttpVersion(1, 1),
-                                    CIMultiDict(), False, False)
+                                    CIMultiDict(), [], False, False)
         self.payload = mock.Mock()
         self.transport = mock.Mock()
         self.reader = mock.Mock()
@@ -47,7 +53,7 @@ class TestUrlDispatcher(unittest.TestCase):
         return handler
 
     def test_system_route(self):
-        route = SystemRoute(201, 'test')
+        route = SystemRoute(HTTPCreated(reason='test'))
         self.assertIsNone(route.match('any'))
         with self.assertRaises(RuntimeError):
             route.url()
@@ -83,7 +89,7 @@ class TestUrlDispatcher(unittest.TestCase):
         route = PlainRoute('GET', handler, 'return', '/handler/to/path')
         self.assertRaises(ValueError, self.router.register_route, route)
 
-        route = PlainRoute('GET', handler, 'test.test:test',
+        route = PlainRoute('GET', handler, 'test.test:test-test',
                            '/handler/to/path')
         self.router.register_route(route)
 
@@ -241,7 +247,7 @@ class TestUrlDispatcher(unittest.TestCase):
     def test_route_plain(self):
         handler = self.make_handler()
         route = self.router.add_route('GET', '/get', handler, name='name')
-        route2 = self.router['name']
+        route2 = next(iter(self.router['name']))
         url = route2.url()
         self.assertEqual('/get', url)
         self.assertIs(route, route2)
@@ -255,7 +261,7 @@ class TestUrlDispatcher(unittest.TestCase):
         route = self.router.add_route('GET', '/get/{name}', handler,
                                       name='name')
 
-        route2 = self.router['name']
+        route2 = next(iter(self.router['name']))
         url = route2.url(parts={'name': 'John'})
         self.assertEqual('/get/John', url)
         self.assertIs(route, route2)
@@ -271,34 +277,34 @@ class TestUrlDispatcher(unittest.TestCase):
         route = self.router.add_static('/st',
                                        os.path.dirname(aiohttp.__file__),
                                        name='static')
-        route2 = self.router['static']
-        url = route2.url(filename='/dir/a.txt')
+        resource = self.router['static']
+        url = resource.url(filename='/dir/a.txt')
         self.assertEqual('/st/dir/a.txt', url)
-        self.assertIs(route, route2)
+        self.assertIs(route, next(iter(resource)))
 
     def test_plain_not_match(self):
         handler = self.make_handler()
         self.router.add_route('GET', '/get/path', handler, name='name')
         route = self.router['name']
-        self.assertIsNone(route.match('/another/path'))
+        self.assertIsNone(route._match('/another/path'))
 
     def test_dynamic_not_match(self):
         handler = self.make_handler()
         self.router.add_route('GET', '/get/{name}', handler, name='name')
         route = self.router['name']
-        self.assertIsNone(route.match('/another/path'))
+        self.assertIsNone(route._match('/another/path'))
 
     def test_static_not_match(self):
         self.router.add_static('/pre', os.path.dirname(aiohttp.__file__),
                                name='name')
         route = self.router['name']
-        self.assertIsNone(route.match('/another/path'))
+        self.assertIsNone(route._route.match('/another/path'))
 
     def test_dynamic_with_trailing_slash(self):
         handler = self.make_handler()
         self.router.add_route('GET', '/get/{name}/', handler, name='name')
         route = self.router['name']
-        self.assertEqual({'name': 'John'}, route.match('/get/John/'))
+        self.assertEqual({'name': 'John'}, route._match('/get/John/'))
 
     def test_len(self):
         handler = self.make_handler()
@@ -321,20 +327,21 @@ class TestUrlDispatcher(unittest.TestCase):
 
     def test_plain_repr(self):
         handler = self.make_handler()
-        self.router.add_route('GET', '/get/path', handler, name='name')
-        self.assertRegex(repr(self.router['name']),
+        route = PlainRoute('GET', handler, 'name', '/get/path')
+        self.assertRegex(repr(route),
                          r"<PlainRoute 'name' \[GET\] /get/path")
 
     def test_dynamic_repr(self):
         handler = self.make_handler()
-        self.router.add_route('GET', '/get/{path}', handler, name='name')
-        self.assertRegex(repr(self.router['name']),
+        route = DynamicRoute('GET', handler, 'name',
+                             'pattern', '/get/{path}')
+        self.assertRegex(repr(route),
                          r"<DynamicRoute 'name' \[GET\] /get/{path}")
 
     def test_static_repr(self):
         self.router.add_static('/get', os.path.dirname(aiohttp.__file__),
                                name='name')
-        self.assertRegex(repr(self.router['name']),
+        self.assertRegex(repr(next(iter(self.router['name']))),
                          r"<StaticRoute 'name' \[GET\] /get/")
 
     def test_static_adds_slash(self):
@@ -435,9 +442,10 @@ class TestUrlDispatcher(unittest.TestCase):
 
             req = self.make_request('GET', '/get/john')
             match_info = yield from self.router.resolve(req)
+            self.assertEqual({'name': 'john'}, match_info)
             self.maxDiff = None
             self.assertRegex(repr(match_info),
-                             "<MatchInfo {'name': 'john'}: <DynamicRoute.+>>")
+                             "<MatchInfo {'name': 'john'}: .+<Dynamic.+>>")
 
         self.loop.run_until_complete(go())
 
@@ -447,7 +455,8 @@ class TestUrlDispatcher(unittest.TestCase):
         def go():
             req = self.make_request('POST', '/path/to')
             match_info = yield from self.router.resolve(req)
-            self.assertEqual("<MatchInfo: not found>", repr(match_info))
+            self.assertEqual("<MatchInfoError 404: Not Found>",
+                             repr(match_info))
 
         self.loop.run_until_complete(go())
 
@@ -463,8 +472,8 @@ class TestUrlDispatcher(unittest.TestCase):
 
             req = self.make_request('PUT', '/path/to')
             match_info = yield from self.router.resolve(req)
-            self.assertEqual("<MatchInfo: method PUT is not allowed "
-                             "(allowed methods: GET, POST>", repr(match_info))
+            self.assertEqual("<MatchInfoError 405: Method Not Allowed>",
+                             repr(match_info))
 
         self.loop.run_until_complete(go())
 
@@ -481,7 +490,7 @@ class TestUrlDispatcher(unittest.TestCase):
         route = self.router.add_route(
             'GET', '/', self.make_handler(), expect_handler=handler)
         self.assertIs(route._expect_handler, handler)
-        self.assertIsInstance(route, PlainRoute)
+        self.assertIsInstance(route, ResourceRoute)
 
     def test_custom_expect_handler_dynamic(self):
 
@@ -492,7 +501,7 @@ class TestUrlDispatcher(unittest.TestCase):
         route = self.router.add_route(
             'GET', '/get/{name}', self.make_handler(), expect_handler=handler)
         self.assertIs(route._expect_handler, handler)
-        self.assertIsInstance(route, DynamicRoute)
+        self.assertIsInstance(route, ResourceRoute)
 
     def test_expect_handler_non_coroutine(self):
 
@@ -642,3 +651,226 @@ class TestUrlDispatcher(unittest.TestCase):
         self.assertIsInstance(self.router.routes(), Sized)
         self.assertIsInstance(self.router.routes(), Iterable)
         self.assertIsInstance(self.router.routes(), Container)
+
+    def fill_named_resources(self):
+        route1 = self.router.add_route('GET', '/plain', self.make_handler(),
+                                       name='route1')
+        route2 = self.router.add_route('GET', '/variable/{name}',
+                                       self.make_handler(), name='route2')
+        route3 = self.router.add_static('/static',
+                                        os.path.dirname(aiohttp.__file__),
+                                        name='route3')
+        return route1.name, route2.name, route3.name
+
+    def test_named_routes_abc(self):
+        self.assertIsInstance(self.router.named_routes(), Mapping)
+        self.assertNotIsInstance(self.router.named_routes(), MutableMapping)
+
+    def test_named_resources_abc(self):
+        self.assertIsInstance(self.router.named_resources(), Mapping)
+        self.assertNotIsInstance(self.router.named_resources(), MutableMapping)
+
+    def test_named_routes(self):
+        self.fill_named_resources()
+
+        with self.assertWarns(DeprecationWarning):
+            self.assertEqual(3, len(self.router.named_routes()))
+
+    def test_named_resources(self):
+        names = self.fill_named_resources()
+
+        self.assertEqual(3, len(self.router.named_resources()))
+
+        for name in names:
+            self.assertIn(name, self.router.named_routes())
+            self.assertIsInstance(self.router.named_routes()[name],
+                                  AbstractResource)
+
+    def test_resource_adapter_not_match(self):
+        route = PlainRoute('GET', lambda req: None, None, '/path')
+        self.router.register_route(route)
+        resource = route.resource
+        self.assertIsNotNone(resource)
+        self.assertIsNone(resource._route.match('/another/path'))
+
+    def test_resource_adapter_resolve_not_math(self):
+        route = PlainRoute('GET', lambda req: None, None, '/path')
+        self.router.register_route(route)
+        resource = route.resource
+        self.assertEqual((None, {'GET'}),
+                         self.loop.run_until_complete(
+                             resource.resolve('GET', '/another/path')))
+
+    def test_resource_adapter_resolve_bad_method(self):
+        route = PlainRoute('POST', lambda req: None, None, '/path')
+        self.router.register_route(route)
+        resource = route.resource
+        self.assertEqual((None, {'POST'}),
+                         self.loop.run_until_complete(
+                         resource.resolve('GET', '/path')))
+
+    def test_resource_adapter_resolve_wildcard(self):
+        route = PlainRoute('*', lambda req: None, None, '/path')
+        self.router.register_route(route)
+        resource = route.resource
+        match_info, allowed = self.loop.run_until_complete(
+            resource.resolve('GET', '/path'))
+        self.assertEqual(allowed, {'*'})  # TODO: expand wildcard
+        self.assertIsNotNone(match_info)
+
+    def test_resource_adapter_iter(self):
+        route = PlainRoute('GET', lambda req: None, None, '/path')
+        self.router.register_route(route)
+        resource = route.resource
+        self.assertEqual(1, len(resource))
+        self.assertEqual([route], list(resource))
+
+    def test_resource_iter(self):
+        resource = self.router.add_resource('/path')
+        r1 = resource.add_route('GET', lambda req: None)
+        r2 = resource.add_route('POST', lambda req: None)
+        self.assertEqual(2, len(resource))
+        self.assertEqual([r1, r2], list(resource))
+
+    def test_deprecate_bare_generators(self):
+        resource = self.router.add_resource('/path')
+
+        def gen(request):
+            yield
+
+        with self.assertWarns(DeprecationWarning):
+            resource.add_route('GET', gen)
+
+    def test_view_route(self):
+        resource = self.router.add_resource('/path')
+
+        route = resource.add_route('GET', View)
+        self.assertIs(View, route.handler)
+
+    def test_resource_route_match(self):
+        resource = self.router.add_resource('/path')
+        route = resource.add_route('GET', lambda req: None)
+        self.assertEqual({}, route.resource._match('/path'))
+
+    def test_plain_route_url(self):
+        route = PlainRoute('GET', lambda req: None, None, '/path')
+        self.router.register_route(route)
+        self.assertEqual('/path?arg=1', route.url(query={'arg': 1}))
+
+    def test_dynamic_route_url(self):
+        route = DynamicRoute('GET', lambda req: None, None,
+                             '<pattern>', '/{path}')
+        self.router.register_route(route)
+        self.assertEqual('/path?arg=1', route.url(parts={'path': 'path'},
+                                                  query={'arg': 1}))
+
+    def test_dynamic_route_match_not_found(self):
+        route = DynamicRoute('GET', lambda req: None, None,
+                             re.compile('/path/(?P<to>.+)'), '/path/{to}')
+        self.router.register_route(route)
+        self.assertEqual(None, route.match('/another/path'))
+
+    def test_dynamic_route_match_found(self):
+        route = DynamicRoute('GET', lambda req: None, None,
+                             re.compile('/path/(?P<to>.+)'), '/path/{to}')
+        self.router.register_route(route)
+        self.assertEqual({'to': 'to'}, route.match('/path/to'))
+
+    def test_deprecate_register_route(self):
+        route = PlainRoute('GET', lambda req: None, None, '/path')
+        with self.assertWarns(DeprecationWarning):
+            self.router.register_route(route)
+
+    def test_error_on_double_route_adding(self):
+        resource = self.router.add_resource('/path')
+
+        resource.add_route('GET', lambda: None)
+        with self.assertRaises(RuntimeError):
+            resource.add_route('GET', lambda: None)
+
+    def test_error_on_adding_route_after_wildcard(self):
+        resource = self.router.add_resource('/path')
+
+        resource.add_route('*', lambda: None)
+        with self.assertRaises(RuntimeError):
+            resource.add_route('GET', lambda: None)
+
+    def test_http_exception_is_none_when_resolved(self):
+        handler = self.make_handler()
+        self.router.add_route('GET', '/', handler)
+        req = self.make_request('GET', '/')
+        info = self.loop.run_until_complete(self.router.resolve(req))
+        self.assertIsNone(info.http_exception)
+
+    def test_http_exception_is_not_none_when_not_resolved(self):
+        handler = self.make_handler()
+        self.router.add_route('GET', '/', handler)
+        req = self.make_request('GET', '/abc')
+        info = self.loop.run_until_complete(self.router.resolve(req))
+        self.assertEqual(info.http_exception.status, 404)
+
+    def test_match_info_get_info_plain(self):
+        handler = self.make_handler()
+        self.router.add_route('GET', '/', handler)
+        req = self.make_request('GET', '/')
+        info = self.loop.run_until_complete(self.router.resolve(req))
+        self.assertEqual(info.get_info(), {'path': '/'})
+
+    def test_match_info_get_info_dynamic(self):
+        handler = self.make_handler()
+        self.router.add_route('GET', '/{a}', handler)
+        req = self.make_request('GET', '/value')
+        info = self.loop.run_until_complete(self.router.resolve(req))
+        self.assertEqual(info.get_info(),
+                         {'pattern': re.compile('^\\/(?P<a>[^{}/]+)$'),
+                          'formatter': '/{a}'})
+
+    def test_resource_adapter_get_info(self):
+        directory = pathlib.Path(aiohttp.__file__).parent
+        route = self.router.add_static('/st', directory)
+        self.assertEqual(route.resource.get_info(), {'directory': directory,
+                                                     'prefix': '/st/'})
+
+    def test_plain_old_style_route_get_info(self):
+        handler = self.make_handler()
+        route = PlainRoute('GET', handler, 'test', '/handler/to/path')
+        self.router.register_route(route)
+        self.assertEqual(route.get_info(), {'path': '/handler/to/path'})
+
+    def test_dynamic_old_style_get_info(self):
+        handler = self.make_handler()
+        route = DynamicRoute('GET', handler, 'name',
+                             '<pattern>', '/get/{path}')
+        self.router.register_route(route)
+        self.assertEqual(route.get_info(), {'formatter': '/get/{path}',
+                                            'pattern': '<pattern>'})
+
+    def test_system_route_get_info(self):
+        handler = self.make_handler()
+        self.router.add_route('GET', '/', handler)
+        req = self.make_request('GET', '/abc')
+        info = self.loop.run_until_complete(self.router.resolve(req))
+        self.assertEqual(info.get_info()['http_exception'].status, 404)
+
+    def fill_resources(self):
+        resource1 = self.router.add_resource('/plain')
+        resource2 = self.router.add_resource('/variable/{name}')
+        return resource1, resource2
+
+    def test_resources_view_len(self):
+        self.fill_resources()
+        self.assertEqual(2, len(self.router.resources()))
+
+    def test_resources_view_iter(self):
+        resources = self.fill_resources()
+        self.assertEqual(list(resources), list(self.router.resources()))
+
+    def test_resources_view_contains(self):
+        resources = self.fill_resources()
+        for resource in resources:
+            self.assertIn(resource, self.router.resources())
+
+    def test_resources_abc(self):
+        self.assertIsInstance(self.router.resources(), Sized)
+        self.assertIsInstance(self.router.resources(), Iterable)
+        self.assertIsInstance(self.router.resources(), Container)
