@@ -41,7 +41,8 @@ class WebFunctionalSetupMixin:
         return port
 
     @asyncio.coroutine
-    def create_server(self, method, path, handler=None, ssl_ctx=None):
+    def create_server(self, method, path, handler=None, ssl_ctx=None,
+                      logger=log.server_logger):
         app = web.Application(loop=self.loop)
         if handler:
             app.router.add_route(method, path, handler)
@@ -49,7 +50,8 @@ class WebFunctionalSetupMixin:
         port = self.find_unused_port()
         self.handler = app.make_handler(
             keep_alive_on=False,
-            access_log=log.access_logger)
+            access_log=log.access_logger,
+            logger=logger)
         srv = yield from self.loop.create_server(
             self.handler, '127.0.0.1', port, ssl=ssl_ctx)
         protocol = "https" if ssl_ctx else "http"
@@ -79,6 +81,7 @@ class TestWebFunctional(WebFunctionalSetupMixin, unittest.TestCase):
         self.loop.run_until_complete(go())
 
     def test_handler_returns_not_response(self):
+        logger = mock.Mock()
 
         @asyncio.coroutine
         def handler(request):
@@ -86,13 +89,14 @@ class TestWebFunctional(WebFunctionalSetupMixin, unittest.TestCase):
 
         @asyncio.coroutine
         def go():
-            _, _, url = yield from self.create_server('GET', '/', handler)
+            _, _, url = yield from self.create_server('GET', '/', handler,
+                                                      logger=logger)
             resp = yield from request('GET', url, loop=self.loop)
             self.assertEqual(500, resp.status)
             resp.close()
 
-        with mock.patch('aiohttp.server.server_logger'):
-            self.loop.run_until_complete(go())
+        self.loop.run_until_complete(go())
+        logger.exception.assert_called_with("Error handling request")
 
     def test_post_form(self):
 
@@ -142,8 +146,11 @@ class TestWebFunctional(WebFunctionalSetupMixin, unittest.TestCase):
         def handler(request):
             data = yield from request.json()
             self.assertEqual(dct, data)
-            data2 = yield from request.json()
+            data2 = yield from request.json(loads=json.loads)
             self.assertEqual(data, data2)
+            with self.assertWarns(DeprecationWarning):
+                data3 = yield from request.json(loader=json.loads)
+            self.assertEqual(data, data3)
             resp = web.Response()
             resp.content_type = 'application/json'
             resp.body = json.dumps(data).encode('utf8')
@@ -357,6 +364,32 @@ class TestWebFunctional(WebFunctionalSetupMixin, unittest.TestCase):
 
         self.loop.run_until_complete(go())
 
+    def test_expect_default_handler_unknown(self):
+        """Test default Expect handler for unknown Expect value.
+
+        A server that does not understand or is unable to comply with any of
+        the expectation values in the Expect field of a request MUST respond
+        with appropriate error status. The server MUST respond with a 417
+        (Expectation Failed) status if any of the expectations cannot be met
+        or, if there are other problems with the request, some other 4xx
+        status.
+
+        http://www.w3.org/Protocols/rfc2616/rfc2616-sec14.html#sec14.20
+        """
+        @asyncio.coroutine
+        def handler(request):
+            yield from request.post()
+            self.fail('Handler should not proceed to this point in case of '
+                      'unknown Expect header')
+
+        @asyncio.coroutine
+        def go():
+            _, _, url = yield from self.create_server('POST', '/', handler)
+            resp = yield from request('POST', url, headers={'Expect': 'SPAM'},
+                                      loop=self.loop)
+            self.assertEqual(417, resp.status)
+        self.loop.run_until_complete(go())
+
     def test_100_continue(self):
         @asyncio.coroutine
         def handler(request):
@@ -520,7 +553,7 @@ class TestWebFunctional(WebFunctionalSetupMixin, unittest.TestCase):
 
         self.loop.run_until_complete(go())
 
-    def test_http10_keep_alive_default(self):
+    def test_http11_keep_alive_default(self):
 
         @asyncio.coroutine
         def handler(request):
@@ -531,12 +564,31 @@ class TestWebFunctional(WebFunctionalSetupMixin, unittest.TestCase):
         def go():
             _, _, url = yield from self.create_server('GET', '/', handler)
             resp = yield from request('GET', url, loop=self.loop,
-                                      version=HttpVersion10)
-            self.assertEqual('close', resp.headers['CONNECTION'])
+                                      version=HttpVersion11)
+            self.assertNotIn('CONNECTION', resp.headers)
             resp.close()
 
         self.loop.run_until_complete(go())
 
+    def test_http10_keep_alive_default(self):
+
+        @asyncio.coroutine
+        def handler(request):
+            yield from request.read()
+            return web.Response(body=b'OK')
+
+        @asyncio.coroutine
+        def go():
+            _, _, url = yield from self.create_server('GET', '/', handler)
+            with ClientSession(loop=self.loop) as session:
+                resp = yield from session.get(url,
+                                              version=HttpVersion10)
+                self.assertEqual(resp.version, HttpVersion10)
+                self.assertEqual('keep-alive', resp.headers['CONNECTION'])
+                resp.close()
+
+        self.loop.run_until_complete(go())
+
     def test_http09_keep_alive_default(self):
 
         @asyncio.coroutine
@@ -551,7 +603,7 @@ class TestWebFunctional(WebFunctionalSetupMixin, unittest.TestCase):
             resp = yield from request('GET', url, loop=self.loop,
                                       headers=headers,
                                       version=HttpVersion(0, 9))
-            self.assertEqual('close', resp.headers['CONNECTION'])
+            self.assertNotIn('CONNECTION', resp.headers)
             resp.close()
 
         self.loop.run_until_complete(go())
@@ -569,7 +621,7 @@ class TestWebFunctional(WebFunctionalSetupMixin, unittest.TestCase):
             headers = {'Connection': 'close'}
             resp = yield from request('GET', url, loop=self.loop,
                                       headers=headers, version=HttpVersion10)
-            self.assertEqual('close', resp.headers['CONNECTION'])
+            self.assertNotIn('CONNECTION', resp.headers)
             resp.close()
 
         self.loop.run_until_complete(go())
@@ -1019,6 +1071,13 @@ class StaticFileMixin(WebFunctionalSetupMixin):
 
         self.loop.run_until_complete(go(here, filename))
 
+    def test_env_nosendfile(self):
+        directory = os.path.dirname(__file__)
+
+        with mock.patch.dict(os.environ, {'AIOHTTP_NOSENDFILE': '1'}):
+            route = web.StaticRoute(None, "/", directory)
+            self.assertEqual(route._sendfile, route._sendfile_fallback)
+
 
 class TestStaticFileSendfileFallback(StaticFileMixin,
                                      unittest.TestCase):
